import 'dart:collection';
import 'dart:convert';

import 'package:contentsquare/contentsquare.dart';
import 'package:contentsquare/src/modules/logger/logger.dart';
import 'package:contentsquare/src/modules/screengraph/compression.dart';
import 'package:contentsquare/src/modules/screengraph/render_object_extension.dart';
import 'package:contentsquare/src/modules/webview_tracking/javascript_event_handler.dart';
 import 'package:flutter/rendering.dart';
import 'package:flutter/widgets.dart';

import '../screengraph/hit_test/get_element_data_of_hit_render_object.dart';
import '../screengraph/hit_test/hit_element_data.dart';
import 'exceptions/exceptions.dart';
import 'logs/logs.dart';
import 'screengraph_retrieval.dart';

/// Widget used to track a WebView.
///
/// To track of a WebView from the webview_flutter package, check the default
/// constructor documentation.
///
/// To track a WebView from another package, check [ContentsquareWebViewTrackerBuilder.custom].
///
/// {@template contentsquare.webview.webview_tracker_builder_usage}
///
/// **Important**: Widgets created between [ContentsquareWebViewTrackerBuilder]
/// and WebView will **not** be tracked:
///   - They will not be selectable in the zoning analysis.
///   - Gestures on those widgets will not be tracked.
/// Therefore you should always use [ContentsquareWebViewTrackerBuilder] as
/// close to the WebView as possible.
/// {@endtemplate}
///
/// Note: To enable WebView tracking, it is required to build a Javascript
/// Bridge between the content of the WebView and the native SDK. To do so,
/// you will have to implement the Contentsquare WebView Javascript Tracking
/// Tag in the web pages called in your app WebViews.
///   - See https://docs.contentsquare.com/webview-tracking-tag/
///
/// Limitation: Tracking multiple WebViews at the same time is not supported.
///
/// {@endtemplate}
class ContentsquareWebViewTrackerBuilder extends StatefulWidget {
  /// Creates a widget which provides a tracker to track a WebView from any
  /// package.
  ///
  /// For the tracking to work, you need to:
  ///   - Enable javascript in the WebView.
  ///   - Register the given [ContentsquareWebViewTracker.javascriptChannels]
  ///     to allow the webview to communicate with the SDK.
  ///   - Use [ContentsquareWebViewTracker.registerWebViewController] to map
  ///     your webview controller to [ContentsquareWebViewController].
  ///     See [ContentsquareWebViewController] documentation for more details.
  ///   - Use [ContentsquareWebViewTracker.initializeWebViewTracking] to start
  ///     the tracking.
  ///     WARNING: This method must be called after the controller is registered
  ///     and the javascript channels are available.
  ///
  /// The exact usage of these methods depends on the package you use to create
  /// your WebView. Reach out to us if you need help.
  ///
  /// Note: If you are trying to track a WebView from the webview_flutter
  /// package, you should use the default constructor instead.
  const ContentsquareWebViewTrackerBuilder.custom({
    Key? key,
    required this.builder,
  }) : super(key: key);

  /// Creates a widget which provides a tracker to track a WebView for the
  /// webview_flutter package.
  ///
  /// For the tracking to work, you need to:
  ///   - Set [WebView.javascriptMode] to [JavascriptMode.unrestricted]
  ///   - Use [ContentsquareFlutterWebViewTracker.javascriptChannels] in [WebView.javascriptChannels].
  ///   - Use [ContentsquareFlutterWebViewTracker.onWebViewCreated] in [WebView.onWebViewCreated].
  ///   - Use [ContentsquareFlutterWebViewTracker.onPageFinished] in [WebView.onPageFinished].
  ///
  /// Example:
  /// ```dart
  /// ContentsquareWebViewTrackerBuilder(
  ///   builder: (context, tracker) {
  ///     return WebView(
  ///       initialUrl: '...',
  ///       javascriptMode: JavascriptMode.unrestricted,
  ///       javascriptChannels: tracker.javascriptChannels,
  ///       onWebViewCreated: tracker.onWebViewCreated,
  ///       onPageFinished: tracker.onPageFinished,
  ///     );
  ///   },
  /// );
  /// ```
  ///
  /// {@macro contentsquare.webview.webview_tracker_builder_usage}
  ContentsquareWebViewTrackerBuilder({
    Key? key,
    required Widget Function(
      BuildContext context,
      ContentsquareFlutterWebViewTracker tracker,
    )
        builder,
  }) : this.custom(
          key: key,
          builder: (context, tracker) {
            return builder(
                context, ContentsquareFlutterWebViewTracker(tracker));
          },
        );

  /// Callback in which the WebView should be built.
  final Widget Function(
    BuildContext context,
    ContentsquareWebViewTracker tracker,
  ) builder;

  @override
  State<ContentsquareWebViewTrackerBuilder> createState() =>
      _a135();
}


class _a135
    extends State<ContentsquareWebViewTrackerBuilder> {
  
  late final a75 = a138(
    a72: a338,
    a71: _a216,
  );

  
  String? _a216() {
    
    
    
    
    if (!mounted) {
      a187.a39(a177());
      return null;
    }

    
    
    
    
    
    
    final a312 = context as Element;

    
    
    
    
    
    
    
    
    
    
    final a313 = a312.renderObject;
    if (a313 == null) {
      a187.a39(a179());
      return null;
    }
    if (a313 is! a137) {
      a187.a35(
        'The type of render object associated with the $ContentsquareWebViewTrackerBuilder widget is ${a313.runtimeType}.\n'
        'This is an error, it should always be a $a137.\n',
      );
      return null;
    }

    
    
    
    
    
    
    
    
    final a314 = a313.a226;
    if (a314 == null) {
      a187.a39(a178());
      return null;
    }

    final a315 = a267(
      a535: a313,
      a536: a140.a232(
        a500: a314.center,
      ),
    );

    return a315.a13(
      a428: (hitElementData) {
        return a193(hitElementData.a67);
      },
      
      
      a427: () => null,
    );
  }

  @override
  void dispose() {
    a75._a223();
    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    return _a139(
      a493: _a136(
        a488: widget.builder(context, a75),
        a74: a75.a23,
      ),
    );
  }
}


class _a136
    extends SingleChildRenderObjectWidget {
  const _a136({
    Key? a487,
    required this.a74,
    required Widget a488,
  }) : super(key: a487, child: a488);

  
  final Future<Map<String, dynamic>> Function() a74;

  @override
  RenderObject createRenderObject(BuildContext context) {
    return a137(a73: a74);
  }
}


class a137 extends RenderProxyBox {
  
  a137({
    required this.a73,
  });

  
  final Future<Map<String, dynamic>> Function() a73;
}


class a138 implements ContentsquareWebViewTracker {
  
  a138({
    required this.a72,
    required this.a71,
  });

  
  final a142 a72;

  
  final String? Function() a71;

  
  ContentsquareWebViewController? _a316;

  @override
  void registerWebViewController(ContentsquareWebViewController controller) {
    _a316 = controller;
  }

  @override
  Future<void> initializeWebViewTracking() async {
    final _a317 = this._a316;

    if (_a317 == null) {
      return a187.a39(
        a156(
          a532: null,
          a533:
              'WebViewController is null, did you forget to use tracker.onWebViewCreated?',
        ),
      );
    }

    
    final a318 = _a317.isChannelAvailable;
    if (a318 != null) {
      final bool a319;
      try {
        a319 =
            await _a218(
          a490: a318,
        );
      } catch (error) {
        return a187.a39(
          a156(
            a532: await _a317.currentUrl?.call(),
            a533:
                'Unexpected error while executing JS code, check the registration of the javascript channels: $error',
          ),
        );
      }
      if (!a319) {
        return a187.a39(
          a156(
            a532: await _a317.currentUrl?.call(),
            a533:
                'JavascriptChannels are not registered, did you forget to use tracker.javascriptChannels?',
          ),
        );
      }
    }

    try {
      await _a317.evaluateJavascript(
        '''
${_a217(a489: _a317.javascriptChannelCall)}
        
// This is needed because [_webViewController.evaluateJavascript] only works on
// js code that returns a value (and the bridge registration does not return a 
// value)
true;
''',
      );
    } catch (error) {
      return a187.a39(
        a156(
          a532: await _a317.currentUrl?.call(),
          a533:
              'Unexpected error while executing JS code to setup the Javascript bridge: $error',
        ),
      );
    }

    return a187.a39(
      a161(
        a534: await _a317.currentUrl?.call(),
      ),
    );
  }

  @override
  Set<JavascriptChannel> get javascriptChannels {
    return ContentsquareJavascriptEventEnum.values
        .map(
          (event) => a147(
            a57: Contentsquare(),
            a56: event,
            a55: a71,
          ).a1,
        )
        .toSet();
  }

  
  String _a217({
    required String Function(String channelName, String data)
        a489,
  }) {
    return '''
window.CSJavascriptBridge = {
  
  ${ContentsquareJavascriptEventEnum.values.map(
              (event) => '${event.name}: ${a147(
                a57: Contentsquare(),
                a56: event,
                a55: a71,
              ).a20(
                a434: a489,
              )}',
            ).join(',\n\n').replaceAll('\n', '\n  ')}
};''';
  }

  
  Future<bool> _a218({
    required Future<bool> Function(String) a490,
  }) async {
    for (final a320 in javascriptChannels) {
      final a321 =
          await a490(a320.name);

      if (!a321) {
        return false;
      }
    }

    return true;
  }

  
  
  
  
  var _a322 = UnmodifiableListView<Stopwatch>([]);

  
  Stopwatch _a219() {
    final a323 = Stopwatch();
    _a322 = UnmodifiableListView(_a322 + [a323]);
    return a323;
  }

  
  void _a220(Stopwatch a491) {
    a491.stop();
    _a322 = UnmodifiableListView(
      _a322.toList()..remove(a491),
    );
  }

  
  @visibleForTesting
  static String get a5 => '''
    if (window.cs_wvt == undefined || window.cs_wvt == null) {
      // Cast on 'null' for jsonDecode to decode it to dart null.
      'null';
    } else {
      window.cs_wvt.push(['serializeWebView']);
    }
  ''';

  
  Future<Map<String, dynamic>> a23() async {
    final a324 = _a316;
    if (a324 == null) {
      throw a175();
    }

    
    
    final a325 = _a219();
    a325.start();

    bool a221() {
      return a325.elapsed >=
          a72.a66;
    }

    
    
    
    
    bool a222() {
      return !a325.isRunning;
    }

    Map<String, dynamic>? a326;
    
    
    
    
    while (!a221() || !a222()) {
      final a327 = await a324.evaluateJavascript(
        a5,
      );
      final Object? a328;
      try {
        final a329 = jsonDecode(a327);
        if (a329 is String) {
          
          
          
          a328 = jsonDecode(a329);
        } else {
          a328 = a329;
        }
      } on FormatException catch (e) {
        a187.a39(
          a172(
            a42: a327,
            a41: e,
          ),
        );
        continue;
      }

      
      
      
      
      
      
      
      final a330 = a328 == null ||
          a328 is num ||
          a328 is Map<String, dynamic>;
      if (!a330) {
        a187.a39(
          a171(
            a43: a328,
          ),
        );
      }

      if (a328 is Map<String, dynamic>) {
        a326 = a328;
        break;
      }

      await Future.delayed(
          a72.a65);
    }

    _a220(a325);

    if (a326 == null) {
      throw a173(
        a40: a72.a66,
      );
    }

    return a326;
  }

  
  void _a223() {
    _a322.forEach(_a220);
  }
}


class _a139 extends InheritedWidget {
  const _a139({
    Key? a492,
    required Widget a493,
  }) : super(key: a492, child: a493);

  @override
  bool updateShouldNotify(oldWidget) => false;
}


extension a182 on Element {
  
  bool get _a224 {
    return widget is ContentsquareWebViewTrackerBuilder;
  }

  
  bool get a225 {
    final a331 =
        getElementForInheritedWidgetOfExactType<_a139>();
    return a331 != null || _a224;
  }
}
